package Classifying;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.core.*;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.StringToWordVector;

import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Random;

public class WebPageClassifier implements Serializable {
    private static final long serialVersionUID = 1L;
    // The root folder of the training data files
    private static final String rootF = "/media/Master/LOS/";
    // Training data
    private Instances m_data = null;
    // if it needs to rebuild the classifier because there have been added new instances to the training data or not
    private boolean m_UpToDate;
    // Filter
    private StringToWordVector m_Filter = new StringToWordVector();
    // Classifier
    private Classifier m_Classifier = null;

    /**
     * Construction of empty training set
     * Adds the classes to the training data
     * @throws Exception
     */
    public WebPageClassifier(Classifier classifier, String[] classifierOptions) throws Exception {
        m_Classifier = classifier;
        m_Classifier.setOptions(classifierOptions);
        String datasetName = "WebpageClassification";

        FastVector attributes = new FastVector(2);
        attributes.addElement(new Attribute("Content", (FastVector) null));

        // Class attributes
        File root = new File(rootF);
        ArrayList<String> classes = new ArrayList<String>();
        for(String f1 : root.list()) {
            File folder1 = new File(rootF + "/" + f1);
            for(String f2 : folder1.list()) {
                File folder2 = new File(rootF + "/" + f1 + "/" + f2);
                for(String folder : folder2.list()) {
                    if(!classes.contains(f2 + " - " + folder)) {
                        classes.add(f2 + " - " + folder);
                    }
                }
            }
        }
        FastVector classValues = new FastVector(classes.size());
        for(String c : classes) {
            classValues.addElement(c);
        }
        attributes.addElement(new Attribute("Class", classValues));

        // Create a dataset with initial capacity of 100, plus set the index of class.
        m_data = new Instances(datasetName, attributes, 100);
        m_data.setClassIndex(m_data.numAttributes() -1);
    }

    /**
     * Updates the model with the given training page.
     * Makes an instance of the data, sets it's class and adds it to m_data.
     * Also sets m_UpToDate to false so that it will rebuild the classifier before classifying new documents.
     * @throws Exception
     */
    public void updateModel(String content, String classValue) throws Exception {
        // Convert content string into an instance
        Instance instance = makeInstance(content, m_data);

        // Add the class value to the instance
        instance.setClassValue(classValue);

        // Add teh instance to the training data
        m_data.add(instance);

        m_UpToDate = false;
    }

    /**
     * sets the options for the filter, i.e.:
     * adds TD/IDF , normalizes vectors based on document length
     * @throws Exception
     */
    public void setFilterOptions() throws Exception {
        // Initialize filter
        //TF_IDF
        m_Filter.setIDFTransform(true);
        m_Filter.setTFTransform(true);
        // Normalize document lengths
        m_Filter.setNormalizeDocLength(new SelectedTag(StringToWordVector.FILTER_NORMALIZE_ALL, StringToWordVector.TAGS_FILTER));
        m_Filter.setInputFormat(m_data);
    }

    /**
     * Method that sets the options of the filter, filters the data with this filter and then build the classififer.
     * Sets m_UpToDate to true so that it will not need to perform this task when it is already done and no new data is added.
     * @throws Exception
     */
    public void buildClassifier() throws Exception {
        setFilterOptions();

        Instances filteredData = Filter.useFilter(m_data, m_Filter);

        m_Classifier.buildClassifier(filteredData);

        m_UpToDate = true;
    }

    /**
     * Classifies a given document
     * @param content Content of the document
     * @param actualClass the class that the document belongs to
     * @throws Exception
     */
    public void classifyContent(String content, String actualClass) throws Exception {
        // Check if a classifier already has been built
        if(m_data.numInstances() == 0) {
            throw new Exception("No classifier available");
        }

        if(!m_UpToDate) {
            System.out.println("Model not up to date, building new classifier");

            buildClassifier();
        }

        Instances testSet = m_data.stringFreeStructure();

        Instance instance = makeInstance(content, testSet);

        // Filter the instance
        m_Filter.input(instance);
        Instance filteredInstance = m_Filter.output();

        // Get the index of the predicted class value
        double predicted = m_Classifier.classifyInstance(filteredInstance);

        double [] dist = m_Classifier.distributionForInstance(filteredInstance);
        ArrayList<MyClass> myClasses = new ArrayList<MyClass>();
        int count = 0;
        for(double d : dist) {
            myClasses.add(new MyClass(count, d));
            count++;
        }
        Collections.sort(myClasses);

        System.out.println("\nContent classified as: " + (int) predicted + ": " +
                m_data.classAttribute().value((int) predicted) + "\nActual class:             " +
                actualClass + "\n    Probability: " + dist[(int) predicted]);

        if(!actualClass.equals(m_data.classAttribute().value((int) predicted))) {
            System.out.println("WRONG! other possibilities:\n");
            count = 0;
            for(MyClass m : myClasses) {
                count++;
                System.out.println('"' + m_data.classAttribute().value(m.getId()) + '"' + " with prob: " + m.getPercent());
                if(m.getPercent() == 0.0 || count == 30) {
                    break;
                    }
                }
            }
    }

    /**
     * Converts the content of a file into an instance
     * @param content The content of the file
     * @param data the data already in the system (m_data)
     * @return an Instance of the content
     */
    public Instance makeInstance(String content, Instances data) {
        Instance instance = new Instance(2);

        Attribute contentAtt = data.attribute("Content");
        instance.setValue(contentAtt, contentAtt.addStringValue(content));

        instance.setDataset(data);

        return instance;
    }

    /**
     *
     * @param root The string for the root folder of the files
     * @param processed if it should look for files that ends with .processed or not
     * @return ArrayList with type FilePathAndClass with all the filepaths
     */
    public ArrayList<FilePathAndClass> getAllFilePathsFromFolderStructure(String root, boolean processed) {
        ArrayList<FilePathAndClass> filePaths = new ArrayList<FilePathAndClass>();

        File _root = new File(root);
        if(_root.exists()) {
            for(String f1 : _root.list()) {
                File folder1 = new File(root + "/" + f1);
                for(String f2 : folder1.list()) {
                    File folder2 = new File(root + "/" + f1 + "/" + f2);
                    for(String f3 : folder2.list()) {
                        File folder3 = new File(root + "/" + f1 + "/" + f2 + "/" + f3);
                        for(String file : folder3.list()) {
                            if(processed) {
                                if(file.endsWith(".processed")) {
                                    filePaths.add(new FilePathAndClass(f2 + " - " + f3, root + "/" + f1 + "/" + f2 + "/" + f3 + "/" + file));
                                }
                            }
                            else {
                                if(!file.endsWith(".processed")) {
                                    filePaths.add(new FilePathAndClass(f2 + " - " + f3, root + "/" + f1 + "/" + f2 + "/" + f3 + "/" + file));
                                }
                            }
                        }
                    }
                }
            }
        }
        return filePaths;
    }

    /**
     * runs a 5 run, 5 fold evaluation of the learned data
     * @throws Exception
     */
    public void performEval() throws Exception {
        int runs = 5;
        int folds = 5;
        
        Classifier classifier = (Classifier) Utils.forName(Classifier.class, m_Classifier.getClass().getName(), m_Classifier.getOptions());
        setFilterOptions();
        Instances filteredData = Filter.useFilter(m_data, m_Filter);

        for(int i = 0; i < runs; i++) {
            int seed = i + 1;
            Random random = new Random(seed);

            Instances randData = new Instances(filteredData);
            randData.randomize(random);
            if(randData.classAttribute().isNominal()) {
                randData.stratify(folds);
            }
            
            Evaluation evaluation = new Evaluation(randData);
            for(int n = 0; n < folds; n++) {
                System.out.println("Fold " + n);
                Instances train = randData.trainCV(folds, n);
                Instances test = randData.testCV(folds, n);
                
                Classifier classifierCopy = Classifier.makeCopy(classifier);
                classifierCopy.buildClassifier(train);
                evaluation.evaluateModel(classifierCopy, test);
            }

            // output evaluation
            System.out.println();
            System.out.println("=== Setup run " + (i + 1) + " ===");
            System.out.println("Classifier: " + classifier.getClass().getName() + " " + Utils.joinOptions(classifier.getOptions()));
            System.out.println("Dataset: " + m_data.relationName());
            System.out.println("Folds: " + folds);
            System.out.println("Seed: " + seed);
            System.out.println();
            System.out.println(evaluation.toSummaryString("=== " + folds + "-fold Cross-validation run " + (i + 1) + "===", false));
            System.out.println();

        }
    }

    /**
     * Gets the time
     * @return String on the format yyyy-MM-dd HH:mm:ss
     */
    public static String now() {
        final String DATE_FORMAT_NOW = "yyyy-MM-dd HH:mm:ss";
        Calendar cal = Calendar.getInstance();
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
        return sdf.format(cal.getTime());
    }

    /**
     * Writes the learned model to a file for later usage
     * @param path The place it should be stored
     * @param modelFileString Name of the file
     * @throws Exception
     */
    public void writeModelToFile(String path, String modelFileString) throws Exception {
        File f = new File(path + modelFileString);
        if(f.exists()) {
            f.renameTo(new File(path + modelFileString + ".old"));
            f.delete();
        }
        FileOutputStream modelOutFile = new FileOutputStream(path + modelFileString);
        ObjectOutputStream modelOutObjectFile = new ObjectOutputStream(modelOutFile);
        modelOutObjectFile.writeObject(this);
        modelOutObjectFile.flush();
        modelOutFile.close();
    }

    /**
     * Creates a string with the classifier used and all it's options
     * @return String
     */
    public String getClassifierString() {
        String options = " options:";
        for(String s : m_Classifier.getOptions()) {
            options += " " + s;
        }
        return m_Classifier.getClass().toString() + options;
    }
}